/*
    XorGramana Copyright 2009 James W. Morris, james@jwm-art.net

    This file is part of XorGramana.

    XorGramana is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    XorGramana is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with XorGramana.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "map.h"
#include "icons.h"

#include <string.h>
#include <stdlib.h>

#include "options.h"
#include "actions.h"
#include "player.h"
#include "game_display.h" /* purely for map_check_solved */

struct xg_map* map=0;

/*  xg_map_validate ensures that any objects that gravitate
    are suspended by *something*. fish and h-bombs should not
    hang in the air, and chickens and v-bombs should not be
    being held back by nothing, waiting to race left. returns
    0 for invalid maps, 1 for valid maps.

    xg_map_validate is called by xg_map_load, which will exit
    on invalid map encounters, ooh-err.
*/
su_t xg_map_validate();


void xg_map_create(const char* title, int width,int height)
{
    int x,y;
    if(!title||width<2||width>MAP_MAX_W
     ||height<2||height>MAP_MAX_H||strlen(title)>MAX_BUF_LEN)
        return;
    if(!(map=malloc(sizeof(struct xg_map))))
        return;
    map->title=0;
    map->data=0;
    if(!(map->title=calloc(strlen(title)+1,sizeof(char)))){
        free(map);
        map=0;
        return;
    }
    strcpy(map->title,title);
    /* map files don't include boundary wall */
    map->width=width+2;
    map->height=height+2;
    if(!(map->data=calloc(map->height,sizeof(map_t*)))){
        xg_map_destroy(map);
        map=0;
        fprintf(stderr,"Failed to allocate for map data\n");
        return;
    }
    for(y=0;y<map->height;y++){
        if(!(map->data[y]=calloc(map->width,sizeof(map_t)))){
            xg_map_destroy(map);
            map=0;
        }
    }
    /* set boundary wall */
    #ifdef MAP_DEBUG
    printf("setting boundary wall (%d x %d)\n",map->width,map->height);
    #endif
    for(x=0;x<map->width;x++)
        map->data[0][x]=map->data[map->height-1][x]=ICON_WALL;
    for(y=1;y<map->height;y++)
        map->data[y][0]=map->data[y][map->width-1]=ICON_WALL;
    map->teleport[0].x=map->teleport[0].y=0;
    map->teleport[1].x=map->teleport[1].y=0;
    for(x=0;x<=MAX_WORDS_TO_SOLVE;x++){
        map->solve_word[x]=0;
        map->solve_unblock_xy[x].x=0;
        map->solve_unblock_xy[x].y=0;
        map->solved[x]=FALSE;
    }
    map->exit.x=map->player[0].x=map->player[1].x=map->width;
    map->exit.y=map->player[0].y=map->player[1].y=map->height;
    map->mask_count=0;
}

void xg_map_destroy()
{
    xy_t y;
    if(map){
        if(map->filename)
            free(map->filename);
        if(map->title)
            free(map->title);
        if(map->data){
            for(y=0;y<map->height;y++)
                if(map->data[y])
                    free(map->data[y]);
            free(map->data);
        }
        for(y=0;y<=MAX_WORDS_TO_SOLVE;y++)
            if(map->solve_word[y])
                free(map->solve_word[y]);
        free(map);
        map=0;
    }
}

char* xg_map_read_name(FILE* fp)
{
    char tmpbuf[80]={0};
    su_t l;
    char* ret=0;
    char* cp=0;
    if(fgets(tmpbuf, 80, fp)!=tmpbuf)
        return 0;
    if((cp=strchr(tmpbuf,'\r')))
        *cp=0;
    else if((cp=strchr(tmpbuf,'\n')))
        *cp=0;
    l=strlen(tmpbuf);
    if(l>=MAPNAME_MAXCHARS)
        l=MAPNAME_MAXCHARS;
    if(!(ret=malloc(l+1)))
        return 0;
    strncpy(ret,tmpbuf,l);
    ret[l]=0;
    return ret;
}

char* xg_map_load_read_name(su_t level)
{
    char* fname=0;
    FILE* fp=0;
    char* ret=0;
    if(!(fname=options_map_filename(level)))
        if(!(fname=options_map_filename(level)))
            return 0;
    if(!(fp=fopen(fname, "r"))){
        free(fname);
        return 0;
    }
    ret=xg_map_read_name(fp);
    free(fname);
    fclose(fp);
    return ret;
}

void xg_map_load(su_t level)
{
    int moves,width,height,tmpx,tmpy,subx,suby; /* int for fscanf %d */
    xy_t    x,y;
    ctr_t   n;
    su_t    t=0;
    FILE    *fp;
    char    buf[MAX_BUF_LEN]={0};
    char    sc;
    char*   fname=0;
    char*   ptr;
    bool    sdone;
    bool    exitdone=FALSE;
    if(map)
        xg_map_destroy(map);
    if(!(fname=options_map_filename(level))){
        if(!(fname=options_map_filename(level)))
            return;
    }
    #ifdef MAP_DEBUG
    printf("Opening map:%s\n",fname);
    #endif
    if (!(fp=fopen(fname,"rb"))){
        fprintf(stderr,"Cannot open map: %s\n",fname);
        free(fname);
        return;
    }
    if(!(fgets(buf,MAX_BUF_LEN,fp))){
        fprintf(stderr,"Cannot read map name from map: %s\n",fname);
        free(fname);
        return;
    }
    for(n=0;n<MAX_BUF_LEN;n++)
        if(buf[n]=='\n'||buf[n]=='\0'){
            buf[n]='\0';
            break;
        }
    if(n==MAX_BUF_LEN)
        buf[n]=0;
    if(options->game){
        moves=2000;
        width=30;
        height=30;
    }
    else{
        if(fscanf(fp,"%d %d %d",&moves,&width,&height)!=3){
            fprintf(stderr,"Cannot read map move count and dimensions from map: %s\n",fname);
            fclose(fp);
            free(fname);
            return;
        }
    }
    #ifdef MAP_DEBUG
    printf("map: '%s' max moves: %d",buf,moves);
    printf("map: '%s' dimensions: w:%d h:%d\n",buf,width,height);
    #endif
    xg_map_create(buf,width,height);
    if(!map){
        fprintf(stderr,"Failed to create map: %s dimensions %d x %d\n",
            fname,width,height);
        fclose(fp);
        free(fname);
        return;
    }
    map->level=level;
    map->max_moves=moves;
    map->filename=fname;
    for(y=1;y<map->height-1;y++){
        do{
            if(!fgets(buf,MAX_BUF_LEN,fp)){
                fprintf(stderr,
                    "Failed reading map data from map: %s, line:%d\n",
                    fname,y);
                fclose(fp);
                xg_map_destroy(map);
                return;
            }
        }while(buf[0]=='\n');
        if(buf[map->width-1]){/* remember map width+=2 for boundary */
            fprintf(stderr,
                "Excess characters in map: %s, map row:%d\n",fname,y);
            fclose(fp);
            xg_map_destroy(map);
            return;
        }
        switch(strlen(buf)<map->width-1){
            case '\n':
            case '\0':
                break;
            default:
                fprintf(stderr,
                    "Too few characters in map: %s, map row:%d\n",fname,y);
                fprintf(stderr,"('%s')\n",map->data[y]);
                fclose(fp);
                xg_map_destroy(map);
                return;
        }
        for(x=1;x<map->width-1;x++){
            if(options->game){
                switch(buf[x-1]){
                    case '2': buf[x-1]=XOR_PLAYER2; break;
                    case '!': buf[x-1]=XOR_FISH;    break;
                    case '<': buf[x-1]=XOR_CHICKEN; break;
                    case 'o': buf[x-1]=XOR_HBOMB;   break;
                    case 'D': buf[x-1]=XOR_DOLL;    break;
                    case 'S': buf[x-1]=XOR_SWITCH;  break;
                    default:
                        break;
                }
            }
            switch((map->data[y][x]=n=mapchar_to_icon(buf[x-1]))){
                case ICON_MASK:
                    map->mask_count++;
                    break;
                case ICON_PLAYER0:
                case ICON_PLAYER1:
                    map->player[n-ICON_PLAYER0].x=x;
                    map->player[n-ICON_PLAYER0].y=y;
                    break;
                case ICON_TELEPORT:
                    if(t<2){
                        map->teleport[t].x=x;
                        map->teleport[t].y=y;
                        t++;
                    }else{
                        fprintf(stderr,
                            "Too many teleports in map %s, map row:%d (%d)\n",
                            fname,y,x);
                        fclose(fp);
                        xg_map_destroy(map);
                        return;
                    }
                    break;
                case ICON_EXIT:
                    #ifdef MAP_DEBUG
                    printf("Found exit at x:%d y:%d\n",x,y);
                    #endif
                    if(exitdone){
                        fprintf(stderr,
                            "WARNING: too many exits in map %s\n",fname);
                        map->data[map->exit.y][map->exit.x]=ICON_FLOOR;
                    }
                    map->exit.x=x;
                    map->exit.y=y;
                    exitdone=TRUE;
                    break;
                #ifdef MAP_DEBUG
                case ICON_MAP:
                    printf("Map piece located @ x:%d y:%d\n",x,y);
                    break;
                #endif
            }
        }
    }
    for(n=0;n<2;n++){
        if(map->player[n].x==map->width||map->player[n].y==map->height){
            fprintf(stderr,"Player %d missing from map %s\n",n,fname);
            fclose(fp);
            xg_map_destroy(map);
            return;
        }
    }
    if(map->exit.x==map->width||map->exit.y==map->height){
        fprintf(stderr,"No exit found in map %s\n",fname);
        fclose(fp);
        xg_map_destroy(map);
        return;
    }
    for(n=0;n<4;n++){
        if(fscanf(fp, "%d %d", &tmpx,&tmpy)<2){
            fprintf(stderr,"Error loading map piece %d data in map %s\n",n,fname);
            fclose(fp);
            xg_map_destroy(map);
            return;
        }
        else{
            map->mappc[n].x=tmpx;
            map->mappc[n].y=tmpy;
        }
        if(map->data[map->mappc[n].y][map->mappc[n].x]!=ICON_MAP){
            fprintf(stderr,"Mismatched map piece %d location data in map %s\n",n,fname);
            fprintf(stderr,"mappc[%d] x:%d y:%d\n",n,map->mappc[n].x,map->mappc[n].y);
            fclose(fp);
            xg_map_destroy(map);
            return;
        }
    }
    if(options->game){
        fclose(fp);
        return;
    }
    n=0;
    sdone=FALSE;
    while(!sdone)
    {
        switch(fscanf(fp,"%s %d %d %c %d %d",buf,&tmpx,&tmpy,&sc,&subx,&suby)){
            case EOF:
            case 0:
                if(n>0){
                    sdone=TRUE;
                    break;
                }
                fprintf(stderr,
                    "Missing solution word in map %s\n",fname);
            case 1:
                fprintf(stderr,
                    "Missing solution word x position in map %s\n",fname);
            case 2:
                fprintf(stderr,
                    "Missing solution word y position in map %s\n",fname);
            case 3:
                fprintf(stderr,
                    "Missing solution word direction in map %s\n",fname);
            case 4:
                fprintf(stderr,
                    "Missing solution word x unblock in map %s\n",fname);
            case 5:
                fprintf(stderr,
                    "Missing solution word y unblock in map %s\n",fname);
                fclose(fp);
                xg_map_destroy(map);
                return;
            default:
            {
                if(n==MAX_WORDS_TO_SOLVE){
                    fprintf(stderr,
                        "WARNING: Too many words to solve in map %s\n",fname);
                    
                    break;
                }
                if(!(map->solve_word[n]=malloc(strlen(buf)+1))){
                    fprintf(stderr,
                        "Map solution word allocation error word %s map %s\n",
                        buf,fname);
                    fclose(fp);
                    xg_map_destroy(map);
                    return;
                }
                ptr=buf;
                x=0;
                while(*ptr){
                    map->solve_word[n][x]=mapchar_to_icon(*ptr);
                    ptr++;
                    x++;
                }
                map->solve_word[n][x]=0;
                map->solve_xy[n].x=tmpx;
                map->solve_xy[n].y=tmpy;
                map->solve_unblock_xy[n].x=subx;
                map->solve_unblock_xy[n].y=suby;
                switch(sc){
                    case 'd':
                        map->solve_dir[n]=MV_DOWN;    break;
                    case 'r':
                        map->solve_dir[n]=MV_RIGHT;   break;
                    default:
                        fprintf(stderr,
                            "Invalid solution word direction spec (%c) in map %s\n",
                            sc,
                            fname);
                        xg_map_destroy(map);
                        return;
                }
                n++;
            }
            break;
        }
    }
    fclose(fp);
    #ifdef MAP_DEBUG
    printf("-----map data-----\n");
    for(y=0;y<map->height;y++){
        printf("data[%2d][]={ ",y);
        for(x=0;x<map->width;x++)
            printf("%2d ",map->data[y][x]);
        printf("}\n");
    }
    #endif
}

void xg_map_init_views(xy_t view_w, xy_t view_h)
{
    int x,y;
    int hx=view_w/2;
    int hy=view_h/2;
    su_t p;
    if(hx<view_w/2.0f)
        hx++;
    if(hy<view_h/2.0f)
        hy++;
    if(!map){
        fprintf(stderr,"xg_map_init_views called for non-existing map\n");
        return;
    }
    for(p=0;p<2;p++){
        x=map->player[p].x-hx;
        y=map->player[p].y-hy;
        if(x<0)
            x=0;
        else if(x+view_w>=map->width)
            x=map->width-view_w;
        if(y<0)
            y=0;
        else if(y+view_h>=map->height)
            y=map->height-view_h;
        map->view[p].x=x;
        map->view[p].y=y;
    }
}

su_t map_get_teleport(xy_t x, xy_t y)
{
    su_t i;
    for(i=0;i<2;i++)
        if(map->teleport[i].x==x&&map->teleport[i].y==y)
            return i;
    return 99;
}

bool map_check_solved()
{
    xy_t x,y,ux,uy;
    su_t n=0;
    map_t* ptr=0;
    bool solved=TRUE;
    bool exit_exists;
    switch(map->data[map->exit.y][map->exit.x]){
        case ICON_EXIT_OPEN:    exit_exists=TRUE;   break;
        case ICON_EXIT:         exit_exists=TRUE;   break;
        default:                exit_exists=FALSE;  break;
    }
    if(options->game){
        if(player.masks_collected==map->mask_count&&exit_exists==TRUE){
            map->data[map->exit.y][map->exit.x]=ICON_EXIT_OPEN;
            return TRUE;
        }
        return FALSE;
    }
    /* was returning if problem solved... but forgot, it is *
     * always possible to un-solve it. mwuh! ha! ha! ha!    */
    while((ptr=map->solve_word[n])){
        x=map->solve_xy[n].x;
        y=map->solve_xy[n].y;
        map->solved[n]=TRUE;
        while(*ptr&&map->solved[n]){
            if(*ptr!=map->data[y][x])
                solved=map->solved[n]=FALSE;
            ptr++;
            switch(map->solve_dir[n]){
                case MV_DOWN:   y++;    break;
                default:        x++;    break;
            }
        }
        ux=map->solve_unblock_xy[n].x;
        uy=map->solve_unblock_xy[n].y;
        if(ux&&uy&&map->solved[n]){
            map->data[uy][ux]=ICON_FLOOR;
            map->solve_unblock_xy[n].x=0;
            map->solve_unblock_xy[n].y=0;
            game_win_display();
            #ifdef MAPDEVMODE
            map_dump_solve_state();
            #endif
        }
        n++;
    }
    if(!exit_exists)
        return solved;
    if(!solved)
        map->data[map->exit.y][map->exit.x]=ICON_EXIT;
    else
        map->data[map->exit.y][map->exit.x]=ICON_EXIT_OPEN;
    return solved;
}

void xg_map_load_error(FILE* fp,char* filename, char* msg)
{
    if(fp)
        fclose(fp);

    if(filename)
        fprintf(stderr,"Error in map!\n\tfile:%s\n\t%s\n",filename,msg);
    else
        fprintf(stderr,"Error in map!\n\t%s\n",msg);
    xg_map_destroy(map);
    exit(1);
}

#ifdef MAPDEVMODE
void map_dump_solve_state()
{
    int n;
    char word[24];
    map_t* ptr;
    char* cptr;
    printf("\n");
    for(n=0;(ptr=map->solve_word[n])!=0;n++){
        cptr=word;
        while(*ptr){
            *cptr=icon_details[*ptr].mapchar;
            ptr++;
            cptr++;
        }
        *cptr='\0';
        printf("word: %s @x,y:%d,%d dir:%s unblock@x,y:%d,%d state:%s\n",
            word,
            map->solve_xy[n].x,
            map->solve_xy[n].y,
            (map->solve_dir[n]==MV_DOWN?"down":"right"),
            map->solve_unblock_xy[n].x,
            map->solve_unblock_xy[n].y,
            (map->solve_unblock_xy[n].x&&map->solve_unblock_xy[n].y?"UNSOLVED":"SOLVED")
            );
    }
}
#endif

bool is_wall(xy_t x, xy_t y)
{
    if(map->data[y][x]==ICON_WALL)
        return TRUE;
    if(map->data[y][x]>=ICON_EDGE_T&&map->data[y][x]<=ICON_EDGE_TBLR)
        return TRUE;
    return FALSE;
}

void xg_map_apply_edging()
{
    xy_t x,y;
    su_t o,i;
    if(!map)
        return;
    for(y=0;y<map->height;y++){
        for(x=0;x<map->width;x++){
            if(is_wall(x,y)){
                o=0;
                if(y>0)
                    if(!is_wall(x,y-1))
                        o|=EDGE_T;
                if(y<map->height-1)
                    if(!is_wall(x,y+1))
                        o|=EDGE_B;
                if(x>0)
                    if(!is_wall(x-1,y))
                        o|=EDGE_L;
                if(x<map->width-1)
                    if(!is_wall(x+1,y))
                        o|=EDGE_R;
                switch(o){
                    case EDGE_T:    i=ICON_EDGE_T;      break;
                    case EDGE_B:    i=ICON_EDGE_B;      break;
                    case EDGE_L:    i=ICON_EDGE_L;      break;
                    case EDGE_R:    i=ICON_EDGE_R;      break;
                    case EDGE_TB:   i=ICON_EDGE_TB;     break;
                    case EDGE_LR:   i=ICON_EDGE_LR;     break;
                    case EDGE_TL:   i=ICON_EDGE_TL;     break;
                    case EDGE_TR:   i=ICON_EDGE_TR;     break;
                    case EDGE_BL:   i=ICON_EDGE_BL;     break;
                    case EDGE_BR:   i=ICON_EDGE_BR;     break;
                    case EDGE_TBL:  i=ICON_EDGE_TBL;    break;
                    case EDGE_TBR:  i=ICON_EDGE_TBR;    break;
                    case EDGE_TLR:  i=ICON_EDGE_TLR;    break;
                    case EDGE_BLR:  i=ICON_EDGE_BLR;    break;
                    case EDGE_TBLR: i=ICON_EDGE_TBLR;   break;
                    default:        i=ICON_WALL;        break;
                }
                map->data[y][x]=i;
            }
        }
    }
}
